from __future__ import annotations

import asyncio
import json
import mmh3

from memphis.exceptions import MemphisError
from memphis.utils import default_error_handler, get_internal_name
from memphis.message import Message
from memphis.partition_generator import PartitionGenerator


class Consumer:
    MAX_BATCH_SIZE = 5000

    def __init__(
        self,
        connection,
        station_name: str,
        consumer_name,
        consumer_group,
        pull_interval_ms: int,
        batch_size: int,
        batch_max_time_to_wait_ms: int,
        max_ack_time_ms: int,
        max_msg_deliveries: int = 2,
        error_callback=None,
        start_consume_from_sequence: int = 1,
        last_messages: int = -1,
        partition_generator: PartitionGenerator = None,
        subscriptions: dict = None,
    ):
        self.connection = connection
        self.station_name = station_name.lower()
        self.internal_station_name = get_internal_name(self.station_name)
        self.consumer_name = consumer_name.lower()
        self.consumer_group = consumer_group.lower()
        self.pull_interval_ms = pull_interval_ms
        self.batch_size = batch_size
        self.batch_max_time_to_wait_ms = batch_max_time_to_wait_ms
        self.max_ack_time_ms = max_ack_time_ms
        self.max_msg_deliveries = max_msg_deliveries
        self.ping_consumer_interval_ms = 30000
        if error_callback is None:
            error_callback = default_error_handler
        self.t_ping = asyncio.create_task(self.__ping_consumer(error_callback))
        self.start_consume_from_sequence = start_consume_from_sequence
        self.last_messages = last_messages
        self.context = {}
        self.dls_messages = []
        self.dls_current_index = 0
        self.dls_callback_func = None
        self.t_dls = asyncio.create_task(self.__consume_dls())
        self.t_consume = None
        self.inner_station_name = get_internal_name(self.station_name)
        self.subscriptions = subscriptions
        self.partition_generator = partition_generator
        self.cached_messages = []
        self.loading_thread = None


    def set_context(self, context):
        """Set a context (dict) that will be passed to each message handler call."""
        self.context = context

    def consume(self, callback, consumer_partition_key: str = None, consumer_partition_number: int = -1):
        """
        This method starts consuming events from the specified station and invokes the provided callback function for each batch of messages received.

        Parameters:
            callback (function): A function that will be called with each batch of messages received. The function should have the following signature:
                - `callback(messages: List[Message], error: Optional[MemphisError], context: Dict) -> Awaitable[None]`
                - `messages`: A list of `Message` objects representing the batch of messages received.
                - `error`: An optional `MemphisError` object if there was an error while consuming the messages.
                - `context`: A dictionary representing the context that was set using the `set_context()` method.
            consumer_partition_key (str): A string that will be used to determine the partition to consume from. If not provided, the consume will work in a Round Robin fashion.
            consumer_partition_number (int): An integer that will be used to determine the partition to consume from. If not provided, the consume will work in a Round Robin fashion.

        Example:
            import asyncio
            from memphis import Memphis

            async def message_handler(messages, error, context):
                if error:
                    print(f"Error occurred: {error}")
                    return

                for message in messages:
                    print(f"Received message: {message}")

            async def main():
                memphis = Memphis()
                await memphis.connect(host='localhost', username='user', password='pass')
                consumer = await memphis.consumer(station_name='my_station', consumer_name='my_consumer', consumer_group='my_group')
                consumer.set_context({'key': 'value'})
                consumer.consume(message_handler)

                # Keep the event loop running
                while True:
                    await asyncio.sleep(1)
            asyncio.run(main())
        """
        self.dls_callback_func = callback
        self.t_consume = asyncio.create_task(self.__consume(callback, partition_key=consumer_partition_key, consumer_partition_number=consumer_partition_number))

    async def __consume(self, callback, partition_key: str = None, consumer_partition_number: int = -1):
        partition_number = 1
        if consumer_partition_number > 0 and partition_key is not None:
            raise MemphisError('Can not use both partition number and partition key')
        elif partition_key is not None:
            partition_number = self.get_partition_from_key(partition_key)
        elif consumer_partition_number > 0:
            self.validate_partition_number(consumer_partition_number, self.inner_station_name)
            partition_number = consumer_partition_number

        while True:
            if self.connection.is_connection_active and self.pull_interval_ms:
                try:
                    if len(self.subscriptions) > 1:
                        if partition_key is None and consumer_partition_number < 1:
                            partition_number = next(self.partition_generator)

                    memphis_messages = []
                    msgs = await self.subscriptions[partition_number].fetch(self.batch_size)

                    for msg in msgs:
                        memphis_messages.append(
                            Message(msg, self.connection, self.consumer_group, self.internal_station_name)
                        )
                    await callback(memphis_messages, None, self.context)
                    await asyncio.sleep(self.pull_interval_ms / 1000)

                except asyncio.TimeoutError:
                    await callback(
                        [], MemphisError("Memphis: TimeoutError"), self.context
                    )
                    continue
                except Exception as e:
                    if self.connection.is_connection_active:
                        raise MemphisError(str(e)) from e
                    return
            else:
                break

    async def __consume_dls(self):
        subject = get_internal_name(self.station_name)
        consumer_group = get_internal_name(self.consumer_group)
        try:
            subscription_name = "$memphis_dls_" + subject + "_" + consumer_group
            self.consumer_dls = await self.connection.broker_manager.subscribe(
                subscription_name, subscription_name
            )
            async for msg in self.consumer_dls.messages:
                index_to_insert = self.dls_current_index
                if index_to_insert >= 10000:
                    index_to_insert %= 10000
                self.dls_messages.insert(
                    index_to_insert, Message(
                        msg, self.connection, self.consumer_group, self.internal_station_name)
                )
                self.dls_current_index += 1
                if self.dls_callback_func != None:
                    await self.dls_callback_func(
                        [Message(msg, self.connection, self.consumer_group, self.internal_station_name)],
                        None,
                        self.context,
                    )
        except Exception as e:
            if self.dls_callback_func != None:
                await self.dls_callback_func([], MemphisError(str(e)), self.context)
                return

    async def fetch(self, batch_size: int = 10, consumer_partition_key: str = None, consumer_partition_number: int = -1, prefetch: bool = False):
        """
        Fetch a batch of messages.

        Returns a list of Message objects. If the connection is
        not active or no messages are recieved before timing out,
        an empty list is returned.

        Example:

            import asyncio
            
            from memphis import Memphis

            async def main(host, username, password, station):
                memphis = Memphis()
                await memphis.connect(host=host,
                                      username=username,
                                      password=password)
            
                consumer = await memphis.consumer(station_name=station,
                                                  consumer_name="test-consumer",
                                                  consumer_group="test-consumer-group")
            
                while True:
                    batch = await consumer.fetch()
                    print("Recieved {} messages".format(len(batch)))
                    for msg in batch:
                        serialized_record = msg.get_data()
                        print("Message:", serialized_record)
            
                await memphis.close()

            if __name__ == '__main__':
                asyncio.run(main(host,
                                 username,
                                 password,
                                 station))
        
        """
        messages = []
        partition_number = 1
        if len(self.subscriptions) > 1:
            if consumer_partition_number > 0 and consumer_partition_key is not None:
                raise MemphisError('Can not use both partition number and partition key')
            elif consumer_partition_key is not None:
                partition_number = self.get_partition_from_key(consumer_partition_key)
            elif consumer_partition_number > 0:
                self.validate_partition_number(consumer_partition_number, self.inner_station_name)
                partition_number = consumer_partition_number
            else:
                partition_number = next(self.partition_generator)


        if prefetch and len(self.cached_messages) > 0:
            if len(self.cached_messages) >= batch_size:
                messages = self.cached_messages[:batch_size]
                self.cached_messages = self.cached_messages[batch_size:]
                number_of_messages_to_prefetch = batch_size * 2 - batch_size  # calculated for clarity
                self.load_messages_to_cache(number_of_messages_to_prefetch, partition_number)
                return messages
            else:
                messages = self.cached_messages
                batch_size -= len(self.cached_messages)
                self.cached_messages = []

        if self.connection.is_connection_active:
            try:
                if batch_size > self.MAX_BATCH_SIZE:
                    raise MemphisError(
                        f"Batch size can not be greater than {self.MAX_BATCH_SIZE}")
                self.batch_size = batch_size
                if len(self.dls_messages) > 0:
                    if len(self.dls_messages) <= batch_size:
                        messages = self.dls_messages
                        self.dls_messages = []
                        self.dls_current_index = 0
                    else:
                        messages = self.dls_messages[0:batch_size]
                        del self.dls_messages[0:batch_size]
                        self.dls_current_index -= len(messages)
                    return messages
                msgs = await self.subscriptions[partition_number].fetch(batch_size)
                for msg in msgs:
                    messages.append(
                        Message(msg, self.connection, self.consumer_group, self.internal_station_name))
                if prefetch:
                    number_of_messages_to_prefetch = batch_size * 2
                    self.load_messages_to_cache(number_of_messages_to_prefetch, partition_number)
                return messages
            except Exception as e:
                if "timeout" not in str(e).lower():
                    raise MemphisError(str(e)) from e

        return messages


    async def __ping_consumer(self, callback):
        while True:
            try:
                await asyncio.sleep(self.ping_consumer_interval_ms / 1000)
                station_inner = get_internal_name(self.station_name)
                consumer_group = get_internal_name(self.consumer_group)
                if self.inner_station_name in self.connection.partition_consumers_updates_data:
                    for p in self.connection.partition_consumers_updates_data[station_inner]["partitions_list"]:
                        stream_name = f"{station_inner}${str(p)}"
                        await self.connection.broker_connection.consumer_info(
                            stream_name, consumer_group, timeout=30
                        )
                else:
                    stream_name = station_inner
                    await self.connection.broker_connection.consumer_info(
                        stream_name, consumer_group, timeout=30
                    )

            except Exception as e:
                if "consumer not found" in str(e) or "stream not found" in str(e):
                    callback(MemphisError(str(e)))

    async def destroy(self, timeout_retries=5):
        """Destroy the consumer."""
        if self.t_consume is not None:
            self.t_consume.cancel()
        if self.t_dls is not None:
            self.t_dls.cancel()
        if self.t_ping is not None:
            self.t_ping.cancel()
        self.pull_interval_ms = None
        try:
            destroy_consumer_req = {
                "name": self.consumer_name,
                "station_name": self.station_name,
                "username": self.connection.username,
                "connection_id": self.connection.connection_id,
                "req_version": 1,
            }
            consumer_name = json.dumps(
                destroy_consumer_req, indent=2).encode("utf-8")
            # pylint: disable=protected-access
            res = await self.connection._request(
                "$memphis_consumer_destructions", consumer_name, 5, timeout_retries
            )
            # pylint: enable=protected-access
            error = res.data.decode("utf-8")
            if error != "" and not "not exist" in error:
                raise MemphisError(error)
            self.dls_messages.clear()
            internal_station_name = get_internal_name(self.station_name)
            if self.connection.schema_updates_data != {}:
                clients_number = (
                    self.connection.clients_per_station.get(
                        internal_station_name) - 1
                )
                self.connection.clients_per_station[
                    internal_station_name
                ] = clients_number

                if clients_number == 0:
                    sub = self.connection.schema_updates_subs.get(
                        internal_station_name)
                    task = self.connection.schema_tasks.get(internal_station_name)
                    if internal_station_name in self.connection.schema_updates_data:
                        del self.connection.schema_updates_data[internal_station_name]
                    if internal_station_name in self.connection.schema_updates_subs:
                        del self.connection.schema_updates_subs[internal_station_name]
                    if internal_station_name in self.connection.schema_tasks:
                        del self.connection.schema_tasks[internal_station_name]
                    if task is not None:
                        task.cancel()
                    if sub is not None:
                        await sub.unsubscribe()

            map_key = internal_station_name + "_" + self.consumer_name.lower()
            del self.connection.consumers_map[map_key]
        except Exception as e:
            raise MemphisError(str(e)) from e

    def get_partition_from_key(self, key):
        try:
            index = mmh3.hash(key, self.connection.SEED, signed=False) % len(self.subscriptions)
            return self.connection.partition_consumers_updates_data[self.inner_station_name]["partitions_list"][index]
        except Exception as e:
            raise e

    def validate_partition_number(self, partition_number, station_name):
        partitions_list = self.connection.partition_consumers_updates_data[station_name]["partitions_list"]
        if partitions_list is not None:
            if partition_number < 1 or partition_number > len(partitions_list):
                raise MemphisError("Partition number is out of range")
            elif partition_number not in partitions_list:
                raise MemphisError(f"Partition {str(partition_number)} does not exist in station {station_name}")
        else:
            raise MemphisError(f"Partition {str(partition_number)} does not exist in station {station_name}")

    def load_messages_to_cache(self, batch_size, partition_number):
        if not self.loading_thread or not self.loading_thread.is_alive():
            asyncio.ensure_future(self.__load_messages(batch_size, partition_number))

    async def __load_messages(self, batch_size, partition_number):
        new_messages = await self.fetch(batch_size, consumer_partition_number=partition_number)
        if new_messages is not None:
            self.cached_messages.extend(new_messages)



class MemphisError(Exception):
    def __init__(self, message):
        message = message.replace("nats", "memphis")
        message = message.replace("NATS", "memphis")
        message = message.replace("Nats", "memphis")
        message = message.replace("NatsError", "MemphisError")
        self.message = message
        if message.startswith("memphis:"):
            super().__init__(self.message)
        else:
            super().__init__("memphis: " + self.message)


class MemphisConnectError(MemphisError):
    pass


class MemphisSchemaError(MemphisError):
    pass


class MemphisHeaderError(MemphisError):
    pass



import json
import graphql
from graphql import parse as parse_graphql
from graphql import validate as validate_graphql
from jsonschema import validate
import google.protobuf.json_format as protobuf_json_format
import fastavro

from memphis.exceptions import MemphisError, MemphisSchemaError
from memphis.utils import get_internal_name


class Station:
    def __init__(self, connection, name: str):
        self.connection = connection
        self.name = name.lower()
        self.internal_station_name = get_internal_name(self.name)

    async def validate_msg(self, message):
        if self.connection.schema_updates_data[self.internal_station_name] != {}:
            schema_type = self.connection.schema_updates_data[
                self.internal_station_name
            ]["type"]
            if schema_type == "protobuf":
                message = self.validate_protobuf(message)
                return message
            if schema_type == "json":
                message = self.validate_json_schema(message)
                return message
            if schema_type == "graphql":
                message = self.validate_graphql(message)
                return message
            if schema_type == "avro":
                message = self.validate_avro_schema(message)
                return message
            if hasattr(message, "SerializeToString"):
                msg_to_send = message.SerializeToString()
                return msg_to_send
        elif isinstance(message, str):
            message = message.encode("utf-8")
            return message
        elif isinstance(message, graphql.language.ast.DocumentNode):
            msg = message
            message = str(msg.loc.source.body)
            message = message.encode("utf-8")
            return message
        elif hasattr(message, "SerializeToString"):
            msg_to_send = message.SerializeToString()
            return msg_to_send
        elif not isinstance(message, bytearray) and not isinstance(message, dict):
            raise MemphisSchemaError("Unsupported message type")
        else:
            if isinstance(message, dict):
                message = bytearray(json.dumps(message).encode("utf-8"))
            return message

    def validate_protobuf(self, message):
        proto_msg = self.connection.proto_msgs[self.internal_station_name]
        msg_to_send = ""
        try:
            if isinstance(message, bytearray):
                msg_to_send = bytes(message)
                try:
                    proto_msg.ParseFromString(msg_to_send)
                    proto_msg.SerializeToString()
                    msg_to_send = msg_to_send.decode("utf-8")
                except Exception as e:
                    if "parsing message" in str(e):
                        e = "Invalid message format, expecting protobuf"
                    raise MemphisSchemaError(str(e))
                return message
            if hasattr(message, "SerializeToString"):
                msg_to_send = message.SerializeToString()
                proto_msg.ParseFromString(msg_to_send)
                proto_msg.SerializeToString()
                try:
                    proto_msg.ParseFromString(msg_to_send)
                    proto_msg.SerializeToString()
                except Exception as e:
                    if "parsing message" in str(e):
                        e = "Error parsing protobuf message"
                    raise MemphisSchemaError(str(e))
                return msg_to_send
            elif isinstance(message, dict):
                try:
                    protobuf_json_format.ParseDict(message, proto_msg)
                    msg_to_send = proto_msg.SerializeToString()
                    return msg_to_send
                except Exception as e:
                    raise MemphisSchemaError(str(e))
            else:
                raise MemphisSchemaError("Unsupported message type")

        except Exception as e:
            raise MemphisSchemaError("Schema validation has failed: " + str(e))

    def validate_json_schema(self, message):
        try:
            if isinstance(message, bytearray):
                try:
                    message_obj = json.loads(message)
                except Exception as e:
                    raise Exception("Expecting Json format: " + str(e))
            elif isinstance(message, dict):
                message_obj = message
                message = bytearray(json.dumps(message_obj).encode("utf-8"))
            else:
                raise Exception("Unsupported message type")

            validate(
                instance=message_obj,
                schema=self.connection.json_schemas[self.internal_station_name],
            )
            return message
        except Exception as e:
            raise MemphisSchemaError("Schema validation has failed: " + str(e))

    def validate_graphql(self, message):
        try:
            if isinstance(message, bytearray):
                msg = message.decode("utf-8")
                msg = parse_graphql(msg)
            elif isinstance(message, str):
                msg = parse_graphql(message)
                message = message.encode("utf-8")
            elif isinstance(message, graphql.language.ast.DocumentNode):
                msg = message
                message = str(msg.loc.source.body)
                message = message.encode("utf-8")
            else:
                raise Exception("Unsupported message type")
            validate_res = validate_graphql(
                schema=self.connection.graphql_schemas[self.internal_station_name],
                document_ast=msg,
            )
            if len(validate_res) > 0:
                raise Exception(
                    "Schema validation has failed: " + str(validate_res))
            return message
        except Exception as e:
            if "Syntax Error" in str(e):
                e = "Invalid message format, expected GraphQL"
            raise MemphisSchemaError("Schema validation has failed: " + str(e))

    def validate_avro_schema(self, message):
        try:
            if isinstance(message, bytearray):
                try:
                    message_obj = json.loads(message)
                except Exception as e:
                    raise Exception("Expecting Avro format: " + str(e))
            elif isinstance(message, dict):
                message_obj = message
                message = bytearray(json.dumps(message_obj).encode("utf-8"))
            else:
                raise Exception("Unsupported message type")

            fastavro.validate(
                message_obj,
                self.connection.avro_schemas[self.internal_station_name],
            )
            return message
        except fastavro.validation.ValidationError as e:
            raise MemphisSchemaError("Schema validation has failed: " + str(e))

    async def destroy(self, timeout_retries=5):
        """Destroy the station."""
        try:
            name_req = {"station_name": self.name, "username": self.connection.username}
            station_name = json.dumps(name_req, indent=2).encode("utf-8")
            # pylint: disable=protected-access
            res = await self.connection._request(
                "$memphis_station_destructions", station_name, 5, timeout_retries
            )
            # pylint: enable=protected-access
            error = res.data.decode("utf-8")
            if error != "" and not "not exist" in error:
                raise MemphisError(error)

            internal_station_name = get_internal_name(self.name)
            sub = self.connection.schema_updates_subs.get(internal_station_name)
            task = self.connection.schema_tasks.get(internal_station_name)
            if internal_station_name in self.connection.schema_updates_data:
                del self.connection.schema_updates_data[internal_station_name]
            if internal_station_name in self.connection.schema_updates_subs:
                del self.connection.schema_updates_subs[internal_station_name]
            if internal_station_name in self.connection.clients_per_station:
                del self.connection.clients_per_station[internal_station_name]
            if internal_station_name in self.connection.schema_tasks:
                del self.connection.schema_tasks[internal_station_name]
            if task is not None:
                task.cancel()
            if sub is not None:
                await sub.unsubscribe()


            if internal_station_name in self.connection.functions_clients_per_station:
                del self.connection.functions_clients_per_station[internal_station_name]
            if internal_station_name in self.connection.functions_updates_data:
                del self.connection.functions_updates_data[internal_station_name]
            if internal_station_name in self.connection.functions_updates_subs:
                function_sub = self.connection.functions_updates_subs.get(internal_station_name)
                if function_sub is not None:
                    await function_sub.unsubscribe()
                del self.connection.functions_updates_subs[internal_station_name]
            if internal_station_name in self.connection.functions_tasks:
                task = self.connection.functions_tasks.get(internal_station_name)
                if task is not None:
                    task.cancel()
                del self.connection.functions_tasks[internal_station_name]

            self.connection.producers_map = {
                k: v
                for k, v in self.connection.producers_map.items()
                if self.name not in k
            }

            self.connection.consumers_map = {
                k: v
                for k, v in self.connection.consumers_map.items()
                if self.name not in k
            }

        except Exception as e:
            raise MemphisError(str(e)) from e



#The PartitionGenerator class is used to create a round robin generator for station's partitions
#the class gets a list of partitions and by using the next() function it returns an item from the list

class PartitionGenerator:
    def __init__(self, partitions_list):
        self.partitions_list = partitions_list
        self.current = 0
        self.length = len(partitions_list)

    def __next__(self):
        partition_to_return = self.partitions_list[self.current]
        self.current = (self.current + 1) % self.length
        return partition_to_return
    


import random
from threading import Timer
from typing import Callable


class set_interval:
    def __init__(self, func: Callable, sec: int):
        def func_wrapper():
            self.t = Timer(sec, func_wrapper)
            self.t.start()
            func()

        self.t = Timer(sec, func_wrapper)
        self.t.start()

    def cancel(self):
        self.t.cancel()


def default_error_handler(e):
    print("ping exception raised", e)


def get_internal_name(name: str) -> str:
    name = name.lower()
    return name.replace(".", "#")


def random_bytes(amount: int) -> str:
    lst = [random.choice("0123456789abcdef") for n in range(amount)]
    s = "".join(lst)
    return s



from __future__ import annotations

import json

from memphis.exceptions import MemphisConnectError, MemphisError, MemphisSchemaError
from memphis.station import Station

class Message:
    def __init__(self, message, connection, cg_name, internal_station_name):
        self.message = message
        self.connection = connection
        self.cg_name = cg_name
        self.internal_station_name = internal_station_name
        self.station = Station(connection, internal_station_name)

    async def ack(self):
        """Ack a message is done processing."""
        try:
            await self.message.ack()
        except Exception as e:
            if (
                "$memphis_pm_id" in self.message.headers
                and "$memphis_pm_cg_name" in self.message.headers
            ):
                try:
                    msg = {
                        "id": int(self.message.headers["$memphis_pm_id"]),
                        "cg_name": self.message.headers["$memphis_pm_cg_name"],
                    }
                    msg_to_ack = json.dumps(msg).encode("utf-8")
                    await self.connection.broker_manager.publish(
                        "$memphis_pm_acks", msg_to_ack
                    )
                except Exception as e_1:
                    raise MemphisConnectError(str(e_1))
            else:
                raise MemphisConnectError(str(e)) from e
            return

    def get_data(self):
        """Receive the message."""
        try:
            return bytearray(self.message.data)
        except Exception:
            return

    async def get_data_deserialized(self):
        """Receive the message."""
        try:
            if self.connection.schema_updates_data and self.connection.schema_updates_data[self.internal_station_name] != {}:
                schema_type = self.connection.schema_updates_data[
                    self.internal_station_name
                ]["type"]
                try:
                    await self.station.validate_msg(bytearray(self.message.data))
                except Exception as e:
                    raise MemphisSchemaError("Deserialization has been failed since the message format does not align with the currently attached schema: " + str(e))
                if schema_type == "protobuf":
                    proto_msg = self.connection.proto_msgs[self.internal_station_name]
                    proto_msg.ParseFromString(self.message.data)
                    return proto_msg
                if schema_type == "avro":
                    return json.loads(bytearray(self.message.data))
                if schema_type == "json":
                    return json.loads(bytearray(self.message.data))
                if schema_type == "graphql":
                    message = bytearray(self.message.data)
                    decoded_str = message.decode("utf-8")
                    return decoded_str
            else:
                return bytearray(self.message.data)
        except Exception:
            return

    def get_headers(self):
        """Receive the headers."""
        try:
            return self.message.headers
        except Exception:
            return

    def get_sequence_number(self):
        """Get message sequence number."""
        try:
            return self.message.metadata.sequence.stream
        except Exception:
            return

    async def delay(self, delay):
        """Delay and resend the message after delay seconds"""
        if (
            "$memphis_pm_id" in self.message.headers
            and "$memphis_pm_cg_name" in self.message.headers
        ):
            raise MemphisError("cannot delay DLS message")
        try:
            await self.message.nak(delay=delay)
        except Exception as e:
            raise MemphisError(str(e)) from e



from __future__ import annotations

import asyncio
import json
import time
from typing import Union, List
import warnings

import mmh3
from memphis.exceptions import MemphisError, MemphisSchemaError
from memphis.headers import Headers
from memphis.utils import get_internal_name
from memphis.partition_generator import PartitionGenerator
from memphis.station import Station

schemaverse_fail_alert_type = "schema_validation_fail_alert"


class Producer:
    def __init__(
        self, connection, producer_name: str, station_name: Union[str, List[str]] , real_name: str
    ):
        self.connection = connection
        self.producer_name = producer_name.lower()
        self.station_name = station_name
        self.real_name = real_name
        self.background_tasks = set()
        if isinstance(station_name, list):
            self.is_multi_station_producer = True
            return
        else:
            self.is_multi_station_producer = False

        self.station = Station(connection, station_name)
        self.internal_station_name = get_internal_name(self.station_name)
        self.loop = asyncio.get_running_loop()
        if self.internal_station_name in connection.partition_producers_updates_data:
            self.partition_generator = PartitionGenerator(connection.partition_producers_updates_data[self.internal_station_name]["partitions_list"])

    # pylint: disable=too-many-arguments
    async def produce(
        self,
        message,
        ack_wait_sec: int = 15,
        headers: Union[Headers, None] = None,
        async_produce: Union[bool, None] = None,
        nonblocking: bool = False,
        msg_id: Union[str, None] = None,
        concurrent_task_limit: Union[int, None] = None,
        producer_partition_key: Union[str, None] = None,
        producer_partition_number: Union[int, -1] = -1
    ):
        """Produces a message into a station.
        Args:
            message (bytearray/dict): message to send into the station
                                      - bytearray/protobuf class
                                        (schema validated station - protobuf)
                                      - bytearray/dict (schema validated station - json schema)
                                      - string/bytearray/graphql.language.ast.DocumentNode
                                        (schema validated station - graphql schema)
                                      - bytearray/dict (schema validated station - avro schema)
            ack_wait_sec (int, optional): max time in seconds to wait for an ack from the broker.
                                          Defaults to 15 sec.
            headers (dict, optional): message headers, defaults to {}.
            async_produce (boolean, optional): produce operation won't block (wait) on message send.
                                               This argument is deprecated. Please use the
                                               `nonblocking` arguemnt instead.
            nonblocking (boolean, optional): produce operation won't block (wait) on message send.
            msg_id (string, optional): Attach msg-id header to the message in order to
                                       achieve idempotency.
            concurrent_task_limit (int, optional): Limit the number of outstanding async produce
                                                   tasks. Calls with nonblocking=True will block
                                                   if the limit is hit and will wait until the
                                                   buffer drains halfway down.
            producer_partition_key (string, optional): Produce messages to a specific partition using the partition key.
            producer_partition_number (int, optional): Produce messages to a specific partition using the partition number.
        Raises:
            Exception: _description_
        """
        if self.is_multi_station_producer:
            await self._multi_station_produce(
                message,
                ack_wait_sec=ack_wait_sec,
                headers=headers,
                async_produce=async_produce,
                msg_id=msg_id,
                producer_partition_key=producer_partition_key,
                producer_partition_number=producer_partition_number
            )
        else:
            await self._single_station_produce(
                message,
                ack_wait_sec=ack_wait_sec,
                headers=headers,
                async_produce=async_produce,
                nonblocking=nonblocking,
                msg_id=msg_id,
                concurrent_task_limit=concurrent_task_limit,
                producer_partition_key=producer_partition_key,
                producer_partition_number=producer_partition_number
            )

    async def _single_station_produce(
        self,
        message,
        ack_wait_sec: int = 15,
        headers: Union[Headers, None] = None,
        async_produce: Union[bool, None] = None,
        nonblocking: bool = False,
        msg_id: Union[str, None] = None,
        concurrent_task_limit: Union[int, None] = None,
        producer_partition_key: Union[str, None] = None,
        producer_partition_number: Union[int, -1] = -1
    ):
        """Produces a message into a station.
        Args:
            message (bytearray/dict): message to send into the station
                                      - bytearray/protobuf class
                                        (schema validated station - protobuf)
                                      - bytearray/dict (schema validated station - json schema)
                                      - string/bytearray/graphql.language.ast.DocumentNode
                                        (schema validated station - graphql schema)
                                      - bytearray/dict (schema validated station - avro schema)
            ack_wait_sec (int, optional): max time in seconds to wait for an ack from the broker.
                                          Defaults to 15 sec.
            headers (dict, optional): message headers, defaults to {}.
            async_produce (boolean, optional): produce operation won't block (wait) on message send.
                                               This argument is deprecated. Please use the
                                               `nonblocking` arguemnt instead.
            nonblocking (boolean, optional): produce operation won't block (wait) on message send.
            msg_id (string, optional): Attach msg-id header to the message in order to
                                       achieve idempotency.
            concurrent_task_limit (int, optional): Limit the number of outstanding async produce
                                                   tasks. Calls with nonblocking=True will block
                                                   if the limit is hit and will wait until the
                                                   buffer drains halfway down.
            producer_partition_key (string, optional): Produce messages to a specific partition using the partition key.
            producer_partition_number (int, optional): Produce messages to a specific partition using the partition number.
        Raises:
            Exception: _description_
        """
        try:
            message = await self.station.validate_msg(message)

            memphis_headers = {
                "$memphis_producedBy": self.producer_name,
                "$memphis_connectionId": self.connection.connection_id,
            }

            if msg_id is not None and msg_id != "":
                memphis_headers["msg-id"] = msg_id

            if headers is not None:
                headers = headers.headers
                headers.update(memphis_headers)
            else:
                headers = memphis_headers

            if self.internal_station_name not in self.connection.partition_producers_updates_data:
                partition_name = self.internal_station_name
            elif len(self.connection.partition_producers_updates_data[self.internal_station_name]['partitions_list']) == 1:
                partition_name = f"{self.internal_station_name}${self.connection.partition_producers_updates_data[self.internal_station_name]['partitions_list'][0]}"
            elif producer_partition_number > 0 and producer_partition_key is not None:
                raise MemphisError('Can not use both partition number and partition key')
            elif producer_partition_key is not None:
                partition_number = self.get_partition_from_key(producer_partition_key)
                partition_name = f"{self.internal_station_name}${str(partition_number)}"
            elif producer_partition_number > 0:
                self.validate_partition_number(producer_partition_number, self.internal_station_name)
                partition_name = f"{self.internal_station_name}${str(producer_partition_number)}"
            else:
                partition_name = f"{self.internal_station_name}${str(next(self.partition_generator))}"

            if self.internal_station_name in self.connection.functions_updates_data:
                partition_number = partition_name.split("$")[1]
                if partition_number in self.connection.functions_updates_data[self.internal_station_name]:
                    full_subject_name = f"{partition_name}.functions.{self.connection.functions_updates_data[self.internal_station_name][partition_number]}"
                else:
                    full_subject_name = f"{partition_name}.final"
            else:
                full_subject_name = f"{partition_name}.final"

            if async_produce:
                nonblocking = True
                warnings.warn("The argument async_produce is deprecated. " + \
                              "Please use the argument nonblocking instead.")

            if nonblocking:
                try:
                    task = self.loop.create_task(
                               self.connection.broker_connection.publish(
                                 full_subject_name,
                                 message,
                                 timeout=ack_wait_sec,
                                 headers=headers,
                               )
                           )
                    self.background_tasks.add(task)
                    task.add_done_callback(self.background_tasks.discard)

                    # block until the number of outstanding async tasks is reduced
                    if concurrent_task_limit is not None and \
                        len(self.background_tasks) >= concurrent_task_limit:
                        desired_size = concurrent_task_limit / 2
                        while len(self.background_tasks) > desired_size:
                            await asyncio.sleep(0.1)

                    await asyncio.sleep(0)
                except Exception as e:
                    raise MemphisError(e)
            else:
                await self.connection.broker_connection.publish(
                    full_subject_name,
                    message,
                    timeout=ack_wait_sec,
                    headers=headers,
                )
        except MemphisSchemaError as e:
            if self.connection.schema_updates_data[self.internal_station_name] != {}:
                msg_to_send = ""
                if hasattr(message, "SerializeToString"):
                    msg_to_send = message.SerializeToString().decode("utf-8")
                elif isinstance(message, bytearray):
                    msg_to_send = str(message, "utf-8")
                else:
                    msg_to_send = str(message)

                if self.connection.station_schemaverse_to_dls[
                    self.internal_station_name
                ]:
                    unix_time = int(time.time())

                    memphis_headers = {
                        "$memphis_producedBy": self.producer_name,
                        "$memphis_connectionId": self.connection.connection_id,
                    }

                    if headers != {} and not headers == None:
                        headers = headers.headers
                        headers.update(memphis_headers)
                    else:
                        headers = memphis_headers

                    msg_to_send_encoded = msg_to_send.encode("utf-8")
                    msg_hex = msg_to_send_encoded.hex()
                    buf = {
                        "station_name": self.internal_station_name,
                        "producer": {
                            "name": self.producer_name,
                            "connection_id": self.connection.connection_id,
                        },
                        "creation_unix": unix_time,
                        "message": {
                            "data": msg_hex,
                            "headers": headers,
                        },
                        "validation_error": str(e)
                    }
                    buf = json.dumps(buf).encode("utf-8")
                    await self.connection.broker_manager.publish("$memphis_schemaverse_dls", buf)

                if self.connection.cluster_configurations.get(
                    "send_notification"
                ):
                    await self.connection.send_notification(
                        "Schema validation has failed",
                        "Station: "
                        + self.station_name
                        + "\nProducer: "
                        + self.producer_name
                        + "\nError:"
                        + str(e),
                        msg_to_send,
                        schemaverse_fail_alert_type,
                    )
            raise e
        except Exception as e: # pylint: disable-next=no-member
            if hasattr(e, "status_code") and e.status_code == "503":
                raise MemphisError(
                    "Produce operation has failed, please check whether Station/Producer still exist"
                )
            raise MemphisError(str(e)) from e

    async def _multi_station_produce(
        self,
        message,
        ack_wait_sec: int = 15,
        headers: Union[Headers, None] = None,
        async_produce: Union[bool, None] = None,
        msg_id: Union[str, None] = None,
        producer_partition_key: Union[str, None] = None,
        producer_partition_number: Union[int, -1] = -1
    ):
        for sn in self.station_name:
            await self.connection.produce(
                sn,
                self.producer_name,
                message,
                ack_wait_sec=ack_wait_sec,
                headers=headers,
                async_produce=async_produce,
                msg_id=msg_id,
                producer_partition_key=producer_partition_key,
                producer_partition_number=producer_partition_number
            )

    # pylint: enable=too-many-arguments
    async def destroy(self, timeout_retries=5):
        """Destroy the producer."""
        if self.is_multi_station_producer:
            await self._destroy_multi_station_producer(timeout_retries=timeout_retries)
        else:
            await self._destroy_single_station_producer(timeout_retries=timeout_retries)

    async def _destroy_single_station_producer(self, timeout_retries=5):
        """Destroy the producer."""
        try:
            # drain buffered async messages
            while len(self.background_tasks) > 0:
                await asyncio.sleep(0.1)

            destroy_producer_req = {
                "name": self.producer_name,
                "station_name": self.station_name,
                "username": self.connection.username,
                "connection_id": self.connection.connection_id,
                "req_version": 1,
            }

            producer_name = json.dumps(destroy_producer_req).encode("utf-8")
            # pylint: disable=protected-access
            res = await self.connection._request(
                "$memphis_producer_destructions", producer_name, 5, timeout_retries
            )
            # pylint: enable=protected-access
            error = res.data.decode("utf-8")
            if error != "" and not "not exist" in error:
                raise Exception(error)

            internal_station_name = get_internal_name(self.station_name)
            clients_number = (
                self.connection.clients_per_station.get(
                    internal_station_name) - 1
            )
            self.connection.clients_per_station[
                internal_station_name
            ] = clients_number

            if clients_number == 0:
                sub = self.connection.schema_updates_subs.get(
                    internal_station_name)
                task = self.connection.schema_tasks.get(internal_station_name)
                if internal_station_name in self.connection.schema_updates_data:
                    del self.connection.schema_updates_data[internal_station_name]
                if internal_station_name in self.connection.schema_updates_subs:
                    del self.connection.schema_updates_subs[internal_station_name]
                if internal_station_name in self.connection.schema_tasks:
                    del self.connection.schema_tasks[internal_station_name]
                if task is not None:
                    task.cancel()
                if sub is not None:
                    await sub.unsubscribe()


            self.connection.functions_clients_per_station[internal_station_name] -= 1
            if self.connection.functions_clients_per_station[internal_station_name] == 0:
                if internal_station_name in self.connection.functions_updates_data:
                    del self.connection.functions_updates_data[internal_station_name]
                if internal_station_name in self.connection.functions_updates_subs:
                    sub = self.connection.functions_updates_subs.get(internal_station_name)
                    if sub is not None:
                        await sub.unsubscribe()
                    del self.connection.functions_updates_subs[internal_station_name]
                if internal_station_name in self.connection.functions_tasks:
                    task = self.connection.functions_tasks.get(internal_station_name)
                    if task is not None:
                        task.cancel()
                    del self.connection.functions_tasks[internal_station_name]


            map_key = internal_station_name + "_" + self.real_name
            del self.connection.producers_map[map_key]

        except Exception as e:
            raise Exception(e)

    async def _destroy_multi_station_producer(self, timeout_retries=5):
        internal_station_name_list = [get_internal_name(station_name) for station_name in self.station_name]
        producer_keys = [f"{internal_station_name}_{self.real_name}" for internal_station_name in internal_station_name_list]
        producers = [self.connection.producers_map.get(producer_key) for producer_key in producer_keys]
        producers = [producer for producer in producers if producer is not None]
        for producer in producers:
            await producer.destroy(timeout_retries)


    def get_partition_from_key(self, key):
        try:
            index = mmh3.hash(key, self.connection.SEED, signed=False) % len(self.connection.partition_producers_updates_data[self.internal_station_name]["partitions_list"])
            return self.connection.partition_producers_updates_data[self.internal_station_name]["partitions_list"][index]
        except Exception as e:
            raise e

    def validate_partition_number(self, partition_number, station_name):
        partitions_list = self.connection.partition_producers_updates_data[station_name]["partitions_list"]
        if partitions_list is not None:
            if partition_number < 1 or partition_number > len(partitions_list):
                raise MemphisError("Partition number is out of range")
            elif partition_number not in partitions_list:
                raise MemphisError(f"Partition {str(partition_number)} does not exist in station {station_name}")
        else:
            raise MemphisError(f"Partition {str(partition_number)} does not exist in station {station_name}")
